#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include <string>

#include <unistd.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include "Profile.h"
#include "Target.h"
#include "algorithm.h"
#include "param.h"
#include "stats.h"

using namespace std;


/*global variables, used as parameters for the program
*gap_o: open gap penalty
*gap_e: extend gap penalty
*alpha: sequence similarity
*beta: structure similarity
*react_c: cutoff for react (react>=cutoff: one stranded; react<cutoff: two stranded;
*pvalue_c: cutoff for pvalue (pvalue<=cutoff will be shown)
*sflag: flag indicate to use shape specific score function or not
*/ 
double gap_o, gap_e, alpha, beta, react_c, pvalue_c;
int sflag;

struct Tscore{      		
    string sid;				//id of the sequence(for multi-fasta)
    int i;          		//start index in the sequence
    int j;          		//end index in the sequence
    double s;          		//alignment score
    double pvalue;      	//pvalue
    vector<int> ALN;     	//alignment: 2->(x,y); 1->(x, -); 0->(-, y)
    Tscore(string id, int i, int j, double s, double pvalue, const vector<int>& ALN);
    static bool cmp(const Tscore& ts1, const Tscore& ts2);
};


Tscore::Tscore(string sid, int i, int j, double s, double pvalue, const vector<int>& ALN) { this->sid=sid; this->i=i; this->j=j; this->s=s; this->pvalue=pvalue; this->ALN=ALN;}
inline bool operator<(const Tscore& lhs, const Tscore& rhs) { return lhs.pvalue==rhs.pvalue? lhs.s > rhs.s : lhs.pvalue<rhs.pvalue; }


bool overlap(int i1, int j1, int i2, int j2, double ratio){
	int max_i=max(i1, i2);
	int min_j=min(j1, j2);	
	return (min_j-max_i > (j1-i1)*ratio) || (min_j-max_i > (j2-i2)*ratio);
}
	
void print_usage(){
    cout<<"ProbeAlign <-o gap_open> <-e gap_extend> <-a alpha> <-b beta> <-r react cutoff> <-p pvalue cutoff> <-s> profile target target_react"<<endl;
}

void getopt_parser(char** argv, int argc, double* gap_o, double* gap_e, double* alpha, double* beta, double* react_c, double* pvalue_c, int* sflag, int* tflag, char* fn1, char* fn2, char* fn3){
    char c=0;
    int len=0;
    //'o' for gap open penalty, 'e' for gap extend penalty, 'a' for alpha, 'b' for beta, 'r' for react cutoff, 'p' for pvalue cutoff
    //'t' for random test, 's' for using shape score function
    while((c=getopt(argc, argv, "o:e:a:b:r:p:ts"))!=-1){
        switch(c){
            case 'o': *gap_o=atof(optarg); break;
            case 'e': *gap_e=atof(optarg); break;
            case 'a': *alpha=atof(optarg); break;
            case 'b': *beta=atof(optarg); break;
            case 'r': *react_c=atof(optarg); break;
            case 'p': *pvalue_c=atof(optarg); break;
            case 's': *sflag=1; break;
            case 't': *tflag=1; break;
            default: print_usage(); exit(EXIT_FAILURE);
        }
    }

    if(argc-optind != 3){
        print_usage();
        exit(EXIT_FAILURE);
    }

	strcpy(fn1, argv[optind]);
	strcpy(fn2, argv[optind+1]);
	strcpy(fn3, argv[optind+2]);
    return;
}


int main(int argc, char* argv[]){   //argv[1]: profile file argv[2]: target sequence file argv[3]: target react file
    ifstream fs;			//use for opening and reading files
    char profile_fn[255], target_fn[255], target_react_fn[255];
    istringstream iss;
    string line;
    string word;
    int lineno;

	int tflag=0;			//random test or not

    string structure;       	//structure in class profile
    double** SMatrix=0;        	//score matrix in class profile
    int plength=0;            	//length of profile

    map<string, string> seq_map;             //sequences in class Target
    string seq;	
    map<string, vector<double> > react_map;          //scores in class Target
    vector<double> react;
    vector<string> id_vec;                    //sequence id list

    Profile* P;
    Target* T;
    vector<Tscore> Tscore_vec;
    vector<Tscore> ret;						//the final result of the program

    int WS;             //actual window size (the length of target may smaller than WINDOW_SIZE)
    char** TBM;         //alignment traceback matrix
    vector<vector<int> > ALN;  //alignments obtained from one traceback
    vector<int> ALN_sp;      	//start point at target for the alignments
    double* sco;               //scores for semi global alignment
    vector<int> sco_sel;       //selected scores in the regions
    
    double bonus;         	//bonus score for an alignment   
    int bonus_ALN_i;             	//the index of the alignment with optimal bonus score in ALNs

    int k;
    double shape, loc, scale;

    gap_o=GAP_OPEN_DEFAULT_PARAM;
    gap_e=GAP_EXTEND_DEFAULT_PARAM;
    alpha=ALPHA_DEFAULT_PARAM;
    beta=BETA_DEFAULT_PARAM;
    react_c=REACT_CUTOFF_DEFAULT_PARAM;
    pvalue_c=PVALUE_CUTOFF_DEFAULT_PARAM;
    sflag=0;

    getopt_parser(argv, argc, &gap_o, &gap_e, &alpha, &beta, &react_c, &pvalue_c, &sflag, &tflag, profile_fn, target_fn, target_react_fn);
	
//_read arguments
    fs.open(profile_fn);
    if(fs.is_open()!=true){
        cout<<"Profile file can not be opened!!"<<endl;
        exit(EXIT_FAILURE);
    }
	
	if(tflag!=1){
		getline(fs, line);
		iss.str(line);
		iss>>shape>>loc>>scale;
		iss.clear();
    }
    lineno=0;
    while(getline(fs, line)){
        if(lineno==0){
            structure=line;
            plength=line.length();
            SMatrix=new double*[plength];
        }
        else{
            SMatrix[lineno-1]=new double[6];
            iss.str(line);
            for(int i=0; i<6; i++){
				iss>>SMatrix[lineno-1][i];
			}
            iss.clear();
        }
        lineno+=1;
    }
    fs.close();
	
    fs.open(target_fn);
    if(fs.is_open()!=true){
        cout<<"Target sequence file can not be opened!!"<<endl;
        exit(EXIT_FAILURE);
    }

	string _sid;
    while(getline(fs, line)){
        if(line[0]=='>'){
            _sid=line.substr(1, line.find_first_of(" \t\n")-1);
            seq_map[_sid]="";
            id_vec.push_back(_sid);
        }
        else{
			seq_map[_sid]+=line;
		}
    }
    fs.close();
    
    fs.open(target_react_fn);
    if(fs.is_open()!=true){
        cout<<"Target react file can not be opened!!"<<endl;
        exit(EXIT_FAILURE);
    }

    int i=0;
    react_map[id_vec[i]]=vector<double>();
    while(getline(fs, line)){
        if(line.empty()){
            i++;
            if(i >= id_vec.size()){
                cout<<"The numbers of sequences in target and target react files are not the same!"<<endl;
                exit(EXIT_FAILURE);
            }
            react_map[id_vec[i]]=vector<double>();
        }
        else{
            react_map[id_vec[i]].push_back(atof(line.c_str()));
        }
    }
    fs.close();

    //consistency check of target and target react files
    if(react_map.size() != seq_map.size()){
        cout<<"The numbers of sequences in target and target react files are not the same!"<<endl;
        exit(EXIT_FAILURE);
    }
    
    for(vector<string>::iterator it=id_vec.begin(); it!=id_vec.end(); ++it){
        if(react_map[*it].size()!=seq_map[*it].size()){
            cout<<"The numbers of bases in the target and target react files of *"<<*it<<"* are not the same!"<<endl;
            exit(EXIT_FAILURE);
        }
    }

//read arguments_

    P=new Profile(structure.c_str(), SMatrix, plength);
    for(vector<string>::iterator it_i=id_vec.begin(); it_i!=id_vec.end(); it_i++){
        seq=seq_map[*it_i];
        react=react_map[*it_i];

        for(int i=0; i<ceil(seq.length()*1./STEP); i++){
            if(i>0 && seq.length()-i*STEP<WINDOW_SIZE-STEP){    //the alignments in the windows have been considered
                break;
            }
            WS=min((unsigned int)WINDOW_SIZE, (unsigned int)seq.length()-i*STEP);		//WS: actual window size
			
            TBM=new char*[plength+1];						//TBM: traceback matrix (plength+1 x WS+1)
            for(int j=0; j<=plength; j++){
                TBM[j]=new char[WS+1];
            }
            
            sco=new double[WS+1];
			
            T=new Target(seq.substr(i*STEP, WS).c_str(), &react[0]+i*STEP, WS);
			
            sco=SG_affine_align(P, T, TBM);
			
			if(tflag!=1){
				int j=1;
				while(j<=WS){
					if(i>0 && j<=WINDOW_SIZE-STEP) {j++; continue;}
					sco_sel.push_back(j+max_sco(sco+j, min(P->length/SELECT_OPT_RATIO, WS-j+1)));
					j+=P->length/SELECT_OPT_RATIO;
				}
			}
			else{
				for(int j=1; j<=WS; j++){
					if(i>0 && j<=WINDOW_SIZE-STEP) continue;
					sco_sel.push_back(j);
				}
			}
			
            for(vector<int>::iterator it_j=sco_sel.begin(); it_j!=sco_sel.end(); it_j++){
                SG_affine_traceback(TBM, plength, *it_j, ALN, ALN_sp);
                bonus=pair_bonus(P, T, ALN, ALN_sp, &bonus_ALN_i);
                if(tflag==1){
					if(i*STEP+*it_j>=P->length)
						cout<<sco[*it_j]+bonus<<endl;
				}
				else{
					Tscore_vec.push_back(Tscore(*it_i, i*STEP+ALN_sp[bonus_ALN_i], i*STEP+*it_j-1, 
											sco[*it_j]+bonus, pvalue_gamma(sco[*it_j]+bonus, shape, loc, scale), 
											ALN[bonus_ALN_i]));
					
				}
				ALN_sp.clear();
				ALN.clear();
            }
            sco_sel.clear();

            delete T;
            for(int j=0; j<=plength; j++){
                delete[] TBM[j];
            }
            delete[] sco;

        }

		if(tflag!=1){
			sort(Tscore_vec.begin(), Tscore_vec.end());
			for(vector<Tscore>::iterator it_j=Tscore_vec.begin(); it_j!=Tscore_vec.end(); ++it_j){
				for(vector<Tscore>::iterator it_k=it_j+1; it_k!=Tscore_vec.end(); ++it_k){
					if(overlap(it_j->i, it_j->j, it_k->i, it_k->j, 0.2)==true ){
						it_k->i=-1;
						it_k->j=-1;
					}
				}
			}

			for(vector<Tscore>::iterator it=Tscore_vec.begin(); it!=Tscore_vec.end(); ++it){
				if(it->i!=-1 && it->j!=-1 && it->pvalue<=pvalue_c){
					ret.push_back(*it);
				}
			}
			Tscore_vec.clear();
		}
        seq.clear();
        react.clear();
    }

	if(tflag!=1){
		sort(ret.begin(), ret.end());
		for(vector<Tscore>::iterator it=ret.begin(); it!=ret.end(); ++it){
			cout<<"*"<<it->sid<<"\t"<<it->i<<","<<it->j<<"\t"<<it->s<<"\t"<<it->pvalue<<endl;
			show_ALN(P, seq_map[it->sid].c_str(), &(react_map[it->sid][0]), it->i, it->ALN);
		}
	}
    
//_clean
    for(int i=0; i<plength; i++){
        delete[] SMatrix[i];
    }

    ret.clear();
    react_map.clear();
    seq_map.clear();
    id_vec.clear();
    delete P;
//clean_
    return 0;
}

